def loadparam(**param):
    if param.get('systemType') is not None:
        systemType = param.get('systemType')
        if type(systemType) != str:
            raise Exception('systemType should be string')
        print('Setting "systemType" to ' +  str(systemType))
    else:
        raise Exception('Please provide system type "systemType = FMO or SB"')
    if param.get('QDmodel') is not None:
        QDmodel = param.get('QDmodel')
        if type(QDmodel) != str:
            raise Exception('QDmodel should be string')
        if QDmodel not in QDmodel:
            raise ValueError('QDmodel should be one of these: "createQDmodel", "useQDmodel"')
        print('MLQD is running with the option QDmodel = ', QDmodel)
    else:
        QDmodel = 'useQDmodel'
        print('The is running with the dafault option QDmodel = ', QDmodel)
    if param.get('QDmodelType') is not None:
        QDmodelType = param.get('QDmodelType')
        if type(QDmodelType) != str:
            raise Exception('QDmodelType should be string')
        print('Setting ML Model Type "QDmodelType" to ' + str(QDmodelType))
    else:
    	QDmodelType = 'OSTL'
    	print('As the ML model type is not provided, the ML-QD is running with the \n dafault option QDmodelType = OSTL"')
    if QDmodel == 'ceateQDmodel' and QDmodelType == 'KRR':
        n_states = None
    else:
        if param.get('n_states') is not None:
            n_states = param.get('n_states')
            if type(n_states) != int:
                raise Exception(' Number of states "n_states" shoul be integer')
            if param.get('systemType') == 'SB': 
                if n_states != 2:
                    raise ValueError('In spin-boson, n_states should be equal to 2')
            print('Setting number of states "n_states" to ' + str(n_states))
        else:
            if param.get('systemType') == 'FMO':
                n_states = 7   # FMO-7
                print('Running with the default number of states (FMO); n_states: 7, please check if it is the case!')
            if param.get('systemType') == 'SB':
                n_states = 2
                print('Running with the default number of states (SB); n_states: 2')
    return systemType, QDmodel, QDmodelType, n_states



import tensorflow.keras as keras
import pickle
import numpy as np
import mlqd.data as data
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Flatten
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Conv1D
from tensorflow.keras.layers import MaxPooling1D

def CNN_optim(Xin: str, 
            Yin: str, 
            QDmodelOut: str,
            epochs: int, 
            patience: int):
    x_train, y_train, x_val, y_val, kernel_choice = data.data(Xin, Yin)
    print('=================================================================')
    f = open('best_param.pkl', 'rb')
    hyper_param = pickle.load(f)
    f.close()
    filter_choice = [10,30,50,70,90,110,130,150,170,190]
    dense_choice = [8,16,32,64,128,256,512]
    lr_choice = [10**-5, 10**-4, 10**-3, 10**-2, 10**-1]
    batch_size = [8,16,32,64,128,256,512]

    Conv1D_0 = filter_choice[hyper_param['Conv1D']] 
    Conv1D_1 = filter_choice[hyper_param['Conv1D_1']]
    Conv1D_2 = filter_choice[hyper_param['Conv1D_2']]
    Dense_0 = dense_choice[hyper_param['Dense']]
    Dense_1 = dense_choice[hyper_param['Dense_1']]
    Dense_2 = dense_choice[hyper_param['Dense_2']]
    Batch_size = batch_size[hyper_param['batch_size']] 
    If_0 = hyper_param['if']
    If_1 = hyper_param['if_1']
    Kernel_0 = kernel_choice[hyper_param['kernel_size']]
    Kernel_1 = kernel_choice[hyper_param['kernel_size_1']]
    Kernel_2 = kernel_choice[hyper_param['kernel_size_2']]
    Lr_rate = lr_choice[hyper_param['learning_rate']]

    model = Sequential()
    model.add(Conv1D(Conv1D_0, kernel_size=Kernel_0, activation ='relu', input_shape=(x_train.shape[1],1)))
    model.add(Conv1D(Conv1D_1, kernel_size=Kernel_1, activation = 'relu', padding='same'))
    if If_0 == 1:
        model.add(Conv1D(Conv1D_2, kernel_size=Kernel_2, activation = 'relu', padding='same'))
    model.add(MaxPooling1D(pool_size=2))
    model.add(Flatten())
    model.add(Dense(Dense_0, activation = 'relu'))
    model.add(Dense(Dense_1, activation = 'relu'))
    if If_1 == 1: 
        model.add(Dense(Dense_2, activation = 'relu'))
    model.add(Dense(y_train.shape[1], activation='linear'))
    adam = keras.optimizers.Adam(learning_rate=Lr_rate)
    print(model.summary())
    model.compile(loss='mse', optimizer=adam)
    print('=================================================================')
    print('cnn.CNN_optim: Running wth EarlyStopping of patience =', patience)
    
    es = keras.callbacks.EarlyStopping(
        monitor="val_loss",
        min_delta=0,
        patience=patience,
        verbose=2,
        mode="min",
        baseline=None,
        restore_best_weights=True)
    
    print('cnn.CNN_optim: Running with batch size =', Batch_size, 'and epochs =', epochs)
    print('=================================================================')
    model.fit(x_train, y_train,
          batch_size=Batch_size,
          epochs=epochs,
          verbose=2,
          validation_data=(x_val, y_val), callbacks=[es]) 
    QDmodelOut =QDmodelOut + '.hdf5'
    model.save(QDmodelOut)
    print('=================================================================')
    print('cnn.CNN_optim: OSTL model is saved as "', QDmodelOut ,'"') 

def OSTL_default(Xin: str, 
            Yin: str, 
            QDmodelOut: str,
            epochs: int,
            patience: int):


    x_train, y_train, x_val, y_val, kernel_choice = data.data(Xin, Yin)
    print('=================================================================')
    print('cnn.OSTL_default: Running wth EarlyStopping of patience =', patience)
    print('cnn.OSTL_default: Running with batch size = 16 and epochs =', epochs)
    print('=================================================================')

    model = Sequential()
    model.add(Conv1D(80, kernel_size=3, activation ='relu', input_shape=(x_train.shape[1],1)))
    model.add(Conv1D(110, kernel_size=3, activation = 'relu', padding='same'))
    model.add(Conv1D(80, kernel_size=3, activation = 'relu', padding='same'))
    model.add(MaxPooling1D(pool_size=2))
    model.add(Flatten())
    model.add(Dense(32, activation = 'relu'))
    model.add(Dense(128, activation = 'relu'))
    model.add(Dense(y_train.shape[1], activation='linear'))
    adam = keras.optimizers.Adam(learning_rate=10**-3)
    print(model.summary())
    model.compile(loss='mse', optimizer=adam)
    es = keras.callbacks.EarlyStopping(
        monitor="val_loss",
        min_delta=0,
        patience=patience,
        verbose=2,
        mode="min",
        baseline=None,
        restore_best_weights=True)
    model.fit(x_train, y_train,
          batch_size=16,
          epochs=epochs,
          verbose=2,
          validation_data=(x_val, y_val), callbacks=[es])

    QDmodelOut = QDmodelOut + '.hdf5'
    model.save(QDmodelOut)
    print('=================================================================')
    print('cnn.OSTL_default: OSTL model is saved as "', QDmodelOut, '"') 
def AIQD_default(Xin: str, 
            Yin: str, 
            QDmodelOut: str,
            epochs: int,
            patience: int):

    x_train, y_train, x_val, y_val, kernel_choice = data.data(Xin, Yin)
    print('=================================================================')
    print('cnn.AIQD_default: Running wth EarlyStopping of patience =', patience)
    print('cnn.AIQD_default: Running with batch size = 64 and epochs =', epochs)
    print('=================================================================')

    model = Sequential()
    model.add(Conv1D(80, kernel_size=3, activation ='relu', input_shape=(x_train.shape[1],1)))
    model.add(Conv1D(60, kernel_size=3, activation = 'relu', padding='same'))
    model.add(Conv1D(50, kernel_size=3, activation = 'relu', padding='same'))
    model.add(MaxPooling1D(pool_size=2))
    model.add(Flatten())
    model.add(Dense(256, activation = 'relu'))
    model.add(Dense(256, activation = 'relu'))
    model.add(Dense(64, activation = 'relu'))
    model.add(Dense(y_train.shape[1], activation='linear'))
    adam = keras.optimizers.Adam(learning_rate=10**-3)
    print(model.summary())
    model.compile(loss='mse', optimizer=adam)
    es = keras.callbacks.EarlyStopping(
        monitor="val_loss",
        min_delta=0,
        patience=patience,
        verbose=2,
        mode="min",
        baseline=None,
        restore_best_weights=True)
    model.fit(x_train, y_train,
          batch_size=64,
          epochs=epochs,
          verbose=2,
          validation_data=(x_val, y_val), callbacks=[es])
    
    QDmodelOut = QDmodelOut + '.hdf5'
    model.save(QDmodelOut)
    print('=================================================================')
    print('cnn.AIQD_default: AIQD model is saved as "', QDmodelOut, '"')



import os 
import re
import sys
import pickle
import random
import numpy as np
import datetime
from pathlib import Path
import mlqd.ml_dyn as mld
import mlqd.lic as lic
import mlqd.plot as plot
import mlqd.hp  as hp
import mlqd.train_ml as train_ml
import mlqd.createQD as createQD
import mlqd.useQD as useQD
import mlqd.essential_param as essparam

fmo_7_init_sites = [1, 6]
fmo_8_init_sites = [1, 6, 8]
QDmodel = ['createQDmodel', 'useQDmodel']
datatype = ['real', 'imag']
class quant_dyn:
    """ Assign values to the parameters"""
    def __init__(self, **param):
        lic.statement()
        if len(param) == 0:
            hp.help()
        else:
            print('=================================================================')
            print("MLQD is started at", datetime.datetime.now())
            print('=================================================================')
            source_path = Path(__file__).resolve()
            self._mlqdDr = source_path.parent
            [self._systemType, self._QDmodel, self._QDmodelType, self._n_states] = essparam.loadparam(**param)   
############################################################################################
            if self._QDmodel == 'createQDmodel':
                [self._prepInput, self._QDmodelOut, self._energyNorm, self._DeltaNorm,
                self._gammaNorm, self._lambNorm, self._tempNorm, self._Xin, self._Yin, 
                self._hyperParam, self._patience, self._OptEpochs, self._TrEpochs, 
                self._max_evals, self._dataPath, self._xlength, self._krrSigma, 
                self._krrLamb, self._numLogf, self._LogCa, self._LogCb,
                self._LogCc, self._LogCd] = createQD.loadparam(self._mlqdDr, 
                                                            self._systemType, 
                                                            self._QDmodelType, 
                                                            **param)
            else:
                self._prepInput = None
############################################################################################
            if self._QDmodel == 'useQDmodel' or (self._QDmodelType == 'AIQD' and self._QDmodel == 'createQDmodel' and self._prepInput == 'True'):
               if param.get('time') is not None:
                   self._time = param.get('time')
                   print('Setting propagation time "time" to ' + str(self._time))
               else:
                   if self._systemType == 'FMO':
                       self._time = 50 # ps 
                       print('Running with the dafault propagation time; time: 50 ps')
                   elif self._systemType == 'SB':
                       self._time = 20  # a. u.
                       print('Running with the the dafault propagation time:', self._time, '(a.u.)')                
                   else:
                       raise Exception('Provide propagation time using parameter "time"')

               if param.get('time_step') is not None:
                   self._time_step = param.get('time_step')
                   print('Setting time_step to ' + str(self._time_step))
               else:
                   if self._systemType == 'FMO':
                       self._time_step = 5.0 # fs
                       print('Running with the dafault time-step: time_step: 5.0 fs ')
                   elif self._systemType == 'SB':
                       self._time_step = 0.05  # (a.u.)
                       print('Running with the dafault time-step:', self._time_step)
            else:
                self._time = None
                self._time_step = None
###############################################################################################
            if self._QDmodel == 'useQDmodel':
                if param.get('QDmodelIn') is not None:
                    self._QDmodelIn = param.get('QDmodelIn')
                    print('Using the trained model "'+ str(self._QDmodelIn) + '" for dynamics prediction')
                else:
                	raise ValueError('please provide a trained model using the option "QDmodelIn"')
                if param.get('QDtrajOut') is not None:
                    self._traj_output_file = param.get('QDtrajOut')
                    if type(self._traj_output_file) != str:
                        raise Exception('QDtrajOut should be string')
                        print('Trajectory will be save as', str(self._traj_output_file))
                else:
                    self._traj_output_file = str(self._systemType) + "_" + str(self._QDmodelType) + "_QD_" + str(random.random()) + ".npy"
                    print('Trajectory will be save as', str(self._traj_output_file))
########    ######################################################################################
            if self._QDmodelType != 'KRR':
                if self._QDmodel == 'useQDmodel':
                    if self._systemType == 'FMO':
                        if param.get('initState') is not None:
                            self._initState = param.get('initState')
                            if type(self._initState) != int:
                                raise Exception('Initial state "initState" shoul be integer')
                            print('Setting initial state "initState" to ' + str(self._initState))
                        else:
                	        self._initState = 1
                	        print('Running with the default value of initial state; initState = 1')
#################################################################################################
                if self._QDmodelType == 'OSTL' or self._QDmodelType == 'AIQD':
                    if self._QDmodel == 'useQDmodel':
                        if param.get('XfileIn') is None:
                            if self._systemType == 'SB':
                                if param.get('energyDiff') is not None:
                                    self._energyDiff = param.get('energyDiff')
                                    print('Setting energy difference between two states "energyDiff" to ' + str(self._energyDiff))
                                else:
                                    self._energyDiff = 0.0
                                    print('As energy difference is not provided, the ML-QD is running with the dafault option energyDiff = 0.0')
                                if param.get('Delta') is not None:
                                    self._Delta = param.get('Delta')
                                    print('Setting tunneling matrix element of the two states "Delta" to ' + str(self._Delta))
                                else:
                                    self._Delta = 1.0
                                    print('As tunneling is not provided, the ML-QD is running with the dafault option Delta = 1.0')
                            else:
                                self._energyDiff = None
                                self._Delta = None
                        
                            if param.get('gamma') is not None:
                                self._gamma = param.get('gamma')
                                print('Setting cutt-off frequency "gamma" to ' + str(self._gamma))
                            else:
                                if self._systemType == 'FMO':
                            	    self._gamma = 500
                            	    print('Running with the default value of cutt-off frequency; gamma = 500')
                                if self._systemType == 'SB':
                            	    self._gamma = 10
                            	    print('Running with the default value of cutt-off frequency; gamma = 10')
                            
                            if param.get('lamb') is not None:
                                self._lamb = param.get('lamb')
                                print('Setting system-bath coupling strength "lambda" to ' + str(self._lamb))
                            else:
                                if self._systemType == 'FMO':
                                    self._lamb = 520
                                    print('Running with the default value of system-bath coupling strength; lamb = 520')
                                if self._systemType == 'SB':
                                    self._lamb = 1.0
                                    print('Running with the default value of system-bath coupling strength; lamb = 0.1')
                            if param.get('temp') is not None:
                                self._temp = param.get('temp')
                                print('Setting temperature (or inverse temperature) value "temp" to ' + str(self._temp))
                            else:
                                if self._systemType == 'FMO':
                                    self._temp = 510
                                    print('Running with the default temperature (or inverse temperature) value temp = 510')
                                if self._systemType == 'SB':
                                    self._temp = 1.0
                                    print('Running with the default temperature (or inverse temperature) value temp = 1.0')
                                print('=================================================================')
                            if param.get('QDmodelIn') is not None:
                                self._QDmodelIn = param.get('QDmodelIn')
                            else:
                                raise ValueError(str(self._QDmodelIn) + '.pkl does not exist')
                            self._name = re.split(r'.hdf5', self._QDmodelIn)[0] + ".pkl"
                            print('Reading normalization constants from', self._name)
                            print('=================================================================')
                            f = open(self._name, 'rb')   # Load normalization parameters
                            norm_param = pickle.load(f)
                            f.close()
                            self._energyNorm = norm_param['energyNorm']
                            self._DeltaNorm = norm_param['DeltaNorm']
                            self._lambNorm = norm_param['lambNorm']
                            self._gammaNorm = norm_param['gammaNorm']
                            self._tempNorm = norm_param['tempNorm']
                            print('Setting energy difference normalizer "energyNorm" to ' + str(self._energyNorm))
                            print('Setting tunneling matrix element normalizer "DeltaNorm" to ' + str(self._DeltaNorm))
                            print('Setting gamma normalizeer "gammaNorm" to ' + str(self._gammaNorm))
                            print('Setting lambda normalizer "lambNormalizer" to ' + str(self._lambNorm))
                            print('Setting temperature (or inverse temperature) normalizer "tempNorm" to ' + str(self._tempNorm))
                    if self._QDmodel == 'useQDmodel':
                        if self._systemType == 'FMO':
                            if self._n_states == 7:
                	            if self._initState not in fmo_7_init_sites:
                	    	        raise ValueError('The initial State initState for FMO should be 1 or 6')
                            if self._n_states == 8:
                                if self._initState not in fmo_8_init_sites:
                                    raise ValueError('The initial State initState for FMO should be 1, 6 or 8')
######################################################################################################
#                                        Run or Train QD
######################################################################################################
            if self._QDmodel == 'useQDmodel':
                if param.get('XfileIn') is not None:
                    X = param.get('XfileIn')
                    if type(X) == str: 
                        if os.path.isfile(X):
                            self._Xin = np.loadtxt(X)
                            self._Xin = np.reshape(self._Xin, (-1, self._Xin.shape[0]))
                            print('Reading from the input file XfileIn "' + str(X) + '" ..........')
                        else:
                            raise ValueError('The input x-file "' + str(X) + '" does not exist')
                    else:
                        self._Xin = X
                        self._Xin = np.array(X)
                        self._Xin = np.reshape(self._Xin, (-1, self._Xin.shape[0]))
                        print('The input is =', X)
                else:
                    if self._QDmodelType == 'KRR':
                        raise ValueError('please the input shot trajectory through XfileIn')
                    if self._QDmodelType == 'OSTL' or self._QDmodelType == 'AIQD': 
                        if self._systemType == 'FMO':
                            X = np.zeros((1,4), dtype = float)
                            X[0,0] = self._initState/10
                            X[0,1] = self._gamma/self._gammaNorm
                            X[0,2] = self._lamb/self._lambNorm
                            X[0,3] = self._temp/self._tempNorm
                        if self._systemType == 'SB':
                            X = np.zeros((1,5), dtype = float)
                            X[0,0] = self._energyDiff/self._energyNorm
                            X[0,1] = self._Delta/self._DeltaNorm
                            X[0,2] = self._gamma/self._gammaNorm
                            X[0,3] = self._lamb/self._lambNorm
                            X[0,4] = self._temp/self._tempNorm
                        self._Xin = X
########    #########################################################################
            if self._QDmodelType == 'KRR':
                if self._QDmodel == 'createQDmodel':
                    if param.get('dataCol') is not None:
                        self._dataCol = param.get('dataCol')
                        if type(self._dataCol) != int: 
                            raise ValueError('dataCol should be integer')
                        print('MLQD is grabbing column #', self._dataCol, 'as was passed through "dataCol"')
                    else:
                        self._dataCol = 1
                        print('Setting data column "dataCol" to default value ', self._dataCol)
                    if param.get('dtype') is not None:
                        self._dtype = param.get('dtype')
                        if type(self._dtype) != str: 
                            raise ValueError('data type "dtype" should be string')
                        print('data type "dtype" is ', self._dtype)
                        if self._dtype not in datatype:
                            raise ValueError('data type "dtype" should be "real" or "imag"')
                    else:
                        self._dtype = 'real'
                        print('Setting data type "dtype" to default type ', self._dtype)
            else:
                self._dataCol = None
                self._dtype = None
            print('=================================================================')
            
            if self._QDmodelType == 'KRR':
                if self._QDmodel == 'useQDmodel':
                    mld.KRR(
                            self._Xin,
                            self._time,
                            self._time_step,
                            self._QDmodelIn,
                            self._traj_output_file
                            )
                if self._QDmodel == 'createQDmodel':
                    train_ml.KRR(
                                self._Xin,
                                self._Yin,
                                self._QDmodelOut,
                                self._prepInput,
                                self._dataCol,
                                self._xlength,
                                self._dtype,
                                self._dataPath,
                                self._hyperParam,
                                self._krrSigma,
                                self._krrLamb
                                )
            if self._QDmodelType == 'OSTL':
                if self._QDmodel == 'useQDmodel':
                    mld.OSTL(
                    	self._Xin,
                    	self._n_states,
                    	self._time,
                    	self._time_step,
                    	self._QDmodelIn,
                        self._systemType,
                        self._traj_output_file,
                        )
                if self._QDmodel == 'createQDmodel':
                    self._name = self._QDmodelOut + ".pkl"
                    norm_param = {'energyNorm': self._energyNorm, 'DeltaNorm': self._DeltaNorm, 
                                'gammaNorm': self._gammaNorm, 'lambNorm': self._lambNorm, 'tempNorm': self._tempNorm}
                    f = open(self._name, "wb")
                    pickle.dump(norm_param, f)
                    f.close()
                    print('Normalization constants are dumped at', self._name)
                    print('=================================================================')
                    train_ml.OSTL(
                                self._Xin,
                                self._Yin, 
                                self._systemType,
                                self._n_states,
                                self._energyNorm,
                                self._DeltaNorm,
                                self._gammaNorm,
                                self._lambNorm, 
                                self._tempNorm,
                                self._dataPath,
                                self._QDmodelOut,
                                self._hyperParam,
                                self._OptEpochs, 
                                self._TrEpochs,
                                self._max_evals,
                                self._patience,
                                self._prepInput)
            if self._QDmodelType == 'AIQD':
                if self._QDmodel == 'useQDmodel':
                    print('Reading Logistic function parameters from', self._name)
                    self._numLogf = norm_param['numLogf']
                    self._LogCa = norm_param['LogCa']
                    self._LogCb = norm_param['LogCb']
                    self._LogCc = norm_param['LogCc']
                    self._LogCd = norm_param['LogCd']
                    print('Setting number of Logistic functions to', self._numLogf)
                    print('Setting "a" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to ' + str(self._LogCa))
                    print('Setting "b" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to ' + str(self._LogCb))
                    print('Setting "c" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to ' + str(self._LogCc))
                    print('Setting "d" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to ' + str(self._LogCd))
                    print('=================================================================')
                    mld.AIQD(
                    	self._Xin,
                    	self._n_states,
                    	self._time,
                    	self._time_step,
                        self._numLogf, 
                        self._LogCa,
                        self._LogCb,
                        self._LogCc,
                        self._LogCd,
                    	self._QDmodelIn,
                        self._systemType,
                        self._traj_output_file 
                        )
                if self._QDmodel == 'createQDmodel':
                    self._name = self._QDmodelOut + ".pkl"
                    norm_param = {'energyNorm': self._energyNorm, 'DeltaNorm': self._DeltaNorm, 
                            'gammaNorm': self._gammaNorm, 'lambNorm': self._lambNorm, 
                            'tempNorm': self._tempNorm, 'numLogf': self._numLogf, 
                            'LogCa': self._LogCa, 'LogCb': self._LogCb,
                            'LogCc': self._LogCc, 'LogCd': self._LogCd}
                    f = open(self._name, "wb")
                    pickle.dump(norm_param, f)
                    f.close()
                    print('=================================================================')
                    print('Normalization constants and constants for Logistic-function are dumped at', self._name) 
                    print('=================================================================')
                    train_ml.AIQD(
                        self._Xin,
                        self._Yin,
                        self._systemType,
                    	self._n_states,
                    	self._time,
                    	self._time_step,
                        self._numLogf, 
                        self._LogCa,
                        self._LogCb,
                        self._LogCc,
                        self._LogCd,
                        self._energyNorm,
                        self._DeltaNorm,
                        self._gammaNorm,
                        self._lambNorm, 
                        self._tempNorm,
                    	self._QDmodelOut,
                        self._hyperParam,
                        self._OptEpochs,
                        self._TrEpochs,
                        self._max_evals,
                        self._patience,
                        self._prepInput,
                        self._dataPath)

########    ########################################################
#                    parameters for plotting 
########    ########################################################
            print('=================================================================')
            if self._QDmodel == 'useQDmodel':
                if param.get('refTraj') is not None:
                    self._refTraj = param.get('refTraj')
                    print('Using the following reference trajectory for plotting:', self._refTraj)
                    if param.get('pltNstates') is not None:
                        self._pltNstates = param.get('pltNstates')
                    else:
                        self._pltNstates = self._n_states # Number of states to be plotted 
                    if param.get('xlim') is not None:
                        self._xlim = param.get('xlim')
                    else:
                        self._xlim = None # plot the full length dynamics
                    if self._QDmodelType == 'KRR':
                        if param.get('dataCol') is not None:
                            self._dataCol = param.get('dataCol')
                            if type(self._dataCol) != int: 
                                raise ValueError('dataCol should be integer')
                            print('For Plotting: MLQD is grabbing column #', self._dataCol, 'as was passed through "dataCol"')
                        else:
                            self._dataCol = 1
                            print('For Plotting: Setting data column "dataCol" to default value ', self._dataCol)
                        if param.get('dtype') is not None:
                            self._dtype = param.get('dtype')
                            if type(self._dtype) != str: 
                                raise ValueError('data type "dtype" should be string')
                            print('For Plotting: data type "dtype" is ', self._dtype)
                            if self._dtype not in datatype:
                                raise ValueError('data type "dtype" should be "real" or "imag"')
                        else:
                            self._dtype = 'real'
                            print('For Plotting: Setting data type "dtype" to default type ', self._dtype)
                    else:
                        self._dataCol = None
                        self._dtype = None
                    plot.plot(self._n_states,
                            self._QDmodelType,
                            self._traj_output_file, 
                            self._refTraj, 
                            self._dataCol, 
                            self._dtype,
                            self._pltNstates,
                            self._xlim)
                else: 
                   print('No reference trajectory was provided, so dynamics is not plotted.',
                           'You can provide reference trajectory with "refTraj"')
            print('=================================================================')
            print("MLQD is ended at", datetime.datetime.now())




import os
import tensorflow.keras as keras
import pickle
import numpy as np
import mlqd.data as data
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Flatten
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import MaxPooling1D
from tensorflow.keras.layers import Conv1D
from hyperopt import fmin, hp, Trials, STATUS_OK, tpe
from sklearn.model_selection import train_test_split

def optimize(Xin: str, 
            Yin: str, 
            epochs: int, 
            max_evals: int):

    x_train, y_train, x_val, y_val, kernel_choice = data.data(Xin, Yin)
    ######################################################
    print('hyperopt_optim: Optimizing the Neural Network with hyperopt library')
    print('hyperopt_optim: Setting Optimizer to Adam and loss to mean square error (mse)')
    print('hyperopt_optim: We do not optimize the activation function and set it equal to Relu')  
    print('hyperopt_optim: Maximum number of evaluations =', max_evals)
    print('hyperopt_optim: Each evaluation runs for ' + str(epochs) + ' epochs')
    print('=================================================================')
    #####################################################
    space = {'Conv1D': hp.choice('Conv1D', [10,30,50,70,90,110,130,150,170,190]),
            'Conv1D_1': hp.choice('Conv1D_1', [10,30,50,70,90,110,130,150,170,190]),
            'Conv1D_2': hp.choice('Conv1D_2', [10,30,50,70,90,110,130,150,170,190]),
            'Dense': hp.choice('Dense', [8, 16,32,64,128,256,512]),
            'Dense_1': hp.choice('Dense_1', [8, 16,32,64,128,256,512]),
            'Dense_2': hp.choice('Dense_2', [8, 16,32,64,128,256,512]),
            'if': hp.choice('if', [{'layers': 'two', }, {'layers': 'three'}]), 
            'if_1': hp.choice('if_1', [{'layers': 'two', }, {'layers': 'three'}]),
            'kernel_size': hp.choice('kernel_size', kernel_choice),
            'kernel_size_1': hp.choice('kernel_size_1', kernel_choice),
            'kernel_size_2': hp.choice('kernel_size_2', kernel_choice),
            'learning_rate': hp.choice('learning_rate', [10**-5, 10**-4, 10**-3, 10**-2, 10**-1]),
            'batch_size': hp.choice('batch_size', [8, 16,32,64,128,256,512]),
            'activation': 'relu'
        } 
    def optimize_model(params):
        model = Sequential()
        model.add(Conv1D(params['Conv1D'], kernel_size=params['kernel_size'], input_shape=(x_train.shape[1],1)))
        model.add(Activation(params['activation']))
        model.add(Conv1D(params['Conv1D_1'], kernel_size=params['kernel_size_1'], padding='same'))
        model.add(Activation(params['activation']))    
        if params['if']['layers'] == 'three':
            model.add(Conv1D(params['Conv1D_2'], kernel_size=params['kernel_size_2'], padding='same'))
            model.add(Activation(params['activation']))
        model.add(MaxPooling1D(pool_size=2))
        model.add(Flatten())
        model.add(Dense(params['Dense']))
        model.add(Activation(params['activation']))
        model.add(Dense(params['Dense_1']))
        model.add(Activation(params['activation']))
        if params['if_1']['layers'] == 'three':
            model.add(Dense(params['Dense_2']))
            model.add(Activation(params['activation']))

        model.add(Dense(y_train.shape[1], activation='linear'))

        adam = keras.optimizers.Adam(learning_rate=params['learning_rate'])
        model.compile(loss='mse', optimizer=adam)
        model.fit(x_train, y_train,
                  batch_size=params['batch_size'],
                  epochs=epochs,
                  verbose=2,
                  validation_data=(x_val, y_val))
        loss= model.evaluate(x_val, y_val, verbose=0)
        return {'loss': loss, 'status': STATUS_OK, 'model': model}

    #if __name__ == '__main__':
    trials = Trials()
    best_run = fmin(optimize_model,
                        space,
                        algo=tpe.suggest,
                        max_evals=max_evals,
                        trials=trials) 
    f = open("best_param.pkl", "wb")
    pickle.dump(best_run, f)
    f.close()



import re
import pickle

def loadparam(*args, **param):
    systemType = args[0]
    if param.get('XfileIn') is None:
        if systemType == 'SB':
            if param.get('energyDiff') is not None:
                energyDiff = param.get('energyDiff')
                print('Setting energy difference between two states "energyDiff" to ' + str(energyDiff))
            else:
                energyDiff = 0.0
                print('As energy difference is not provided, the ML-QD is running with the dafault option energyDiff = 0.0')
            if param.get('Delta') is not None:
                Delta = param.get('Delta')
                print('Setting tunneling matrix element of the two states "Delta" to ' + str(Delta))
            else:
                Delta = 1.0
                print('As tunneling is not provided, the ML-QD is running with the dafault option Delta = 1.0')
        else:
            energyDiff = None
            Delta = None
    
        if param.get('gamma') is not None:
            gamma = param.get('gamma')
            print('Setting cutt-off frequency "gamma" to ' + str(gamma))
        else:
            if systemType == 'FMO':
        	    gamma = 500
        	    print('Running with the default value of cutt-off frequency; gamma = 500')
            if systemType == 'SB':
        	    gamma = 10
        	    print('Running with the default value of cutt-off frequency; gamma = 10')
        
        if param.get('lamb') is not None:
            lamb = param.get('lamb')
            print('Setting system-bath coupling strength "lambda" to ' + str(lamb))
        else:
            if systemType == 'FMO':
                lamb = 520
                print('Running with the default value of system-bath coupling strength; lamb = 520')
            if systemType == 'SB':
                lamb = 1.0
                print('Running with the default value of system-bath coupling strength; lamb = 0.1')
        
    
        if param.get('temp') is not None:
            temp = param.get('temp')
            print('Setting temperature (or inverse temperature) value "temp" to ' + str(temp))
        else:
            if systemType == 'FMO':
                temp = 510
                print('Running with the default temperature (or inverse temperature) value temp = 510')
            if systemType == 'SB':
                temp = 1.0
                print('Running with the default temperature (or inverse temperature) value temp = 1.0')
        print('=================================================================')
        if param.get('QDmodelIn') is not None:
            QDmodelIn = param.get('QDmodelIn')
        else:
            raise ValueError(str(QDmodelIn) + '.pkl does not exist')
        name = re.split(r'.hdf5', QDmodelIn)[0] + ".pkl"
        print('Reading normalization constants from', name)
        print('=================================================================')
        f = open(name, 'rb')   # Load normalization parameters
        norm_param = pickle.load(f)
        f.close()
        energyNorm = norm_param['energyNorm']
        DeltaNorm = norm_param['DeltaNorm']
        lambNorm = norm_param['lambNorm']
        gammaNorm = norm_param['gammaNorm']
        tempNorm = norm_param['tempNorm']
        print('Setting energy difference normalizer "energyNorm" to ' + str(energyNorm))
        print('Setting tunneling matrix element normalizer "DeltaNorm" to ' + str(DeltaNorm))
        print('Setting gamma normalizeer "gammaNorm" to ' + str(gammaNorm))
        print('Setting lambda normalizer "lambNormalizer" to ' + str(lambNorm))
        print('Setting temperature (or inverse temperature) normalizer "tempNorm" to ' + str(tempNorm))
    else:
        energyDiff = None
        Delta = None
        gamma = None
        lamb = None
        temp = None
        energyNorm = None
        DeltaNorm = None
        lambNorm = None
        gammaNorm = None
        tempNorm = None
    return energyDiff, Delta, gamma, lamb, temp, energyNorm, DeltaNorm, gammaNorm, lambNorm, tempNorm 



import os
import random 

def loadparam(*args, **param):
    mlqdDr = args[0]
    systemType = args[1]
    QDmodelType = args[2]
    if param.get('prepInput') is not None:
        prepInput = param.get('prepInput')
        if type(prepInput) != str:
            raise Exception('prepInput should be string')
        if prepInput == 'True':
            print('Setting option "prepInput" to ' + str(prepInput))
        else:
            print('You have chosen not to prepare the input files, othewise you should pass "True" to prepInput')
    else:
        prepInput = 'False'
        print('The is running with default "prepInput" option', prepInput)
    if param.get('QDmodelOut') is not None:
        QDmodelOut = param.get('QDmodelOut')
        if type(QDmodelOut) != str:
            raise Exception('QDmodelOut should be string')
        print('Model will be saved as ' + str(QDmodelOut))
    else:
        QDmodelOut = str(QDmodelType) + "_" + "model_for_" + str(systemType) + "_" + str(random.random())
        print('The model will be saved as', QDmodelOut)
    if param.get('XfileIn') is not None:
        Xin = param.get('XfileIn')
        if type(Xin) != str:
            raise ValueError('The input XfileIn "' + str(Xin) + '" should be string')
        print('Xfilein is', Xin)
    else:
        Xin = "x_data_" + str(random.random()) 
        print('Setting XfileIn to dafault name', Xin)
    if param.get('YfileIn') is not None:
        Yin = param.get('YfileIn')
        if type(Yin) != str: 
           raise ValueError('The target YfileIn "' + str(Yin) + '" should be string')
        print('YfileIn is', Yin)
    else:
        Yin = "y_data_" + str(random.random())
        print('Setting Yfilein to dafault name', Yin)
    if param.get('hyperParam') is not None:
        hyperParam = param.get('hyperParam')
        if type(hyperParam) != str: 
            raise ValueError('You can only pass string to "hyperParam"')
        if hyperParam == 'True':
            print('You have chosen to optimize the hyper parameters of the model')
        else:
            print('You have chosen not to optimize the hyper parameters of the model, otherwise you should pass "True" to hyperParam')
    else:
        hyperParam = 'False'  # donot optimize
        print('You have chosen not to optimize the hyper parameters of the model, so it will run with the dafault hyper parameters')
    print('=================================================================')
    if QDmodelType == 'OSTL' or QDmodelType == 'AIQD':
        if param.get('patience') is not None:
            patience = param.get('patience')
            if type(patience) != int: 
                raise ValueError('"patience" value can be only integer')
            print('Setting patience for early stopping to', patience)
        else:
            patience = 20  # for early stopping
            print('Running with the dafualt value of "patience" =', patience)
    
        if param.get('OptEpochs') is not None:
            OptEpochs = param.get('OptEpochs')
            if type(OptEpochs) != int: 
                raise ValueError('"OptEpochs" value can be only integer')
            print('Setting number of epochs in hyperopt optimization to', OptEpochs)
        else:
            OptEpochs = 100  # for CNN training or optimization 
            print('Running with the dafualt value of "Optimization epochs" =', OptEpochs)
        if param.get('TrEpochs') is not None:
            TrEpochs = param.get('TrEpochs')
            if type(TrEpochs) != int: 
                raise ValueError('"TrEpochs" value can be only integer')
            print('Setting number of epochs for training to', TrEpochs)
        else:
            TrEpochs = 100  # for CNN training or optimization 
            print('Running with the dafualt value of "Training epochs" =', TrEpochs)
        if hyperParam == 'True':
            if param.get('max_evals') is not None:
                max_evals = param.get('max_evals')
                if type(max_evals) != int: 
                    raise ValueError('"max_evals" value can be only integer')
                print('Setting maximum number of evaluations to', max_evals)
            else:
                max_evals = 100  # for hyperopt optimization 
                print('Running with the dafualt value of maximum evaluations "max_evals" =', max_evals)
        else:
            max_evals = None
        if systemType == 'SB':
            if param.get('energyNorm') is not None:
                energyNorm = param.get('energyNorm')
                print('Setting energy difference normalizer "energyNorm" to ' + str(energyNorm))
            else:
        	    energyNorm = 1.0
        	    print('Running with the default value of energy difference normalizer; energyNorm = 1.0 ')
            if param.get('DeltaNorm') is not None:
                DeltaNorm = param.get('DeltaNorm')
                print('Setting tunneling matrix element normalizer "DeltaNorm" to ' + str(DeltaNorm))
            else:
        	    DeltaNorm = 1.0
        	    print('Running with the default value of tunneling matrix element normalizer; DeltaNorm = 1.0 ')
        else:
            energyNorm = None
            DeltaNorm = None
        if param.get('gammaNorm') is not None:
            gammaNorm = param.get('gammaNorm')
            print('Setting gamma normalizeer "gammaNorm" to ' + str(gammaNorm))
        else:
            if systemType == 'FMO':
        	    gammaNorm = 500.0
        	    print('Running with the default value of gamma normalizeer; gammaNorm = 500 ')
            if systemType == 'SB':
        	    gammaNorm = 10.0
        	    print('Running with the default value of gamma normalizeer; gammaNorm = 10 ')
        if param.get('lambNorm') is not None:
            lambNorm = param.get('lambNorm')
            print('Setting lambda normalizer "lambNormalizer" to ' + str(lambNorm))
        else:
            if systemType == 'FMO':
        	    lambNorm = 510.0
        	    print('Running with the default value of lambda normalizer; lambNorm = 520 ')
            if systemType == 'SB':
        	    lambNorm = 1.0
        	    print('Running with the default value of lambda normalizer; lambNorm = 1.0 ')
        if param.get('tempNorm') is not None:
            tempNorm = param.get('tempNorm')
            print('Setting temperature (or inverse temperature) normalizer "tempNorm" to ' + str(tempNorm))
        else:
            if systemType == 'FMO':
        	    tempNorm = 510.0
        	    print('Running with the default value of temperature (or inverse temperature) normalizer; tempNorm = 510 ')
            if systemType == 'SB':
        	    tempNorm = 1.0
        	    print('Running with the default value of temperature (or inverse temperature) normalizer; tempNorm = 1.0 ')
        print('=================================================================')
    else:
        patience = None
        OptEpochs = None
        TrEpochs = None
        max_evals = None
        energyNorm = None
        DeltaNorm = None
        gammaNorm = None
        lambNorm = None
        tempNorm = None
    if prepInput == 'True':
        if param.get('dataPath') is not None:
             dataPath = param.get('dataPath')
             if type(dataPath) != str: 
                 raise ValueError('The provided datapath "' + str(dataPath) + '" should be string')
             if os.path.isdir(dataPath):
                 pass
             else:
                 raise ValueError('Datapath "' + dataPath +'" does not exist')
        else:
            if systemType == 'SB':
                dataPath = str(mlqdDr) + '/Jupyter_Notebooks/sb_data'
            if systemType == 'FMO':
                dataPath = str(mlqdDr) + '/Jupyter_Notebooks/fmo_data'
    else:
        dataPath = None
###############################################################################
#                                 Specific to KRR
###############################################################################
    if QDmodelType == 'KRR':
        if param.get('xlength') is not None:
            xlength = param.get('xlength')
            if type(xlength) != int: 
                raise ValueError('length of x-input should be integer')
            print('Setting length of each row in input file "xlength" to', xlength)
        else:
            xlength = 81
            print('Setting length of x-input "xlength" to default value', xlength)
    
        if hyperParam != 'True':
            if param.get('krrSigma') is not None:
                krrSigma = param.get('krrSigma')
                print('Setting hyperparameter Sigma for Guassian kernel "krrSigma" to ', krrSigma)
            else:
                krrSigma = 4.0 
                print('Setting hyperparameter Sigma for Gaussian kernel "krrSigma" to default value', krrSigma)
            if param.get('krrLamb') is not None:
                krrLamb = param.get('krrLamb')
                print('Setting hyperparameter Lambda for KRR "krrLamb" to ', krrLamb)
            else:
                krrLamb = 0.00000000047875 
                print('Setting hyperparameter Lambda for KRR "krrLamb" to default value', krrLamb)
        else:
            krrSigma = None
            krrLamb = None
        print('=================================================================')
    else:
        xlength = None
        krrSigma = None
        krrLamb = None
    if QDmodelType == 'AIQD':
        if param.get('numLogf') is not None:
            numLogf = int(param.get('numLogf'))
            if type(numLogf) !=int:
                raise ValueError('numLogf should be integer')
            print('Setting number of Logistic functions to ' + str(numLogf))
        else:
            numLogf = 1
            print('Setting number of Logistic functions to its default value " numLogf = 1 "')
        if param.get('LogCa') is not None:
            LogCa = float(param.get('LogCa'))
            print('Setting "a" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to ' + str(LogCa))
        else:
            LogCa = 1
            print('Setting "a" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to its default value "1"')
        if param.get('LogCb') is not None:
            LogCb = float(param.get('LogCb'))
            print('Setting "b" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to ' + str(LogCb))
        else:
            LogCb = 15
            print('Setting "b" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to its default value "15"')
        if param.get('LogCc') is not None:
            LogCc = float(param.get('LogCc'))
            print('Setting "c" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to ' + str(LogCc))
        else:
            LogCc = -1.0
            print('Setting "b" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to its default value "-1"')
        if param.get('LogCd') is not None:
            LogCd = float(param.get('LogCd'))
            print('Setting "c" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to ' + str(LogCd))
        else:
            LogCd = 1.0
            print('Setting "d" in Logistic function f(t) = a/(1 + b * np.exp(-(t-c)/d))) to its default value "1"')
    else:
        numLogf = None
        LogCa = None
        LogCb = None
        LogCc = None
        LogCd = None

    return prepInput, QDmodelOut, energyNorm, DeltaNorm, gammaNorm, lambNorm, tempNorm, Xin, Yin, hyperParam, patience, OptEpochs, TrEpochs, max_evals, dataPath, xlength, krrSigma, krrLamb, numLogf, LogCa, LogCb, LogCc, LogCd 




"""%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

                Statement

  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
"""

__intro__= "MLQD is a python package developed for Machine Learning-based Quantum Dissipative Dynamics,\n \t\t\t\t  Version 1.1.1\n\t\t\t https://github.com/Arif-PhyChem/MLQD" 
__author__ = "\n \t\t\t Copyright (c) 2023 Arif Ullah"
__copyright__ = "\nAll rights reserved. This work is licensed under the Apache Software License 2.0"
__mod__= "\n\nThe above copyright notice and this permission notice shall be included \nin all copies or substantial portions of the Software."
__disclaimer__="\n\nThe software is provided \"as is\" without warranty of any kind, express or implied, \nincluding but not limited to the warranties ofmerchantability, fitness for a particular \npurpose and noninfringement. In no event shall the authors or copyright holders be \nliable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, \narising from, out of or in connection with the software or the use or other dealings in the software."
__cite__="\n\n\t\t\t\t Cite as:\n\n1) Ullah A. and Dral P. O., New Journal of Physics, 2021, 23(11), 113019\n2) Ullah A. and Dral P. O., Nature Communications, 2022, 13(1), 1930\n3) Ullah A. and Dral P. O., Journal of Physical Chemistry Letters, 2022, 13(26), 6037\n4) Rodriguez L. E. H.; Ullah A.; Espinosa K. J. R.; Dral P. O. and A. A. Kananenka, Machine Learning: Science and Technology, 2022, 3(4), 045016\n5), Ullah A. and Dral P. O., 2023, arXiv.2303.01264"
__contributers__="\nContributers List:\n\n1) Arif Ullah (main) \n2) Pavlo O. Dral"
def statement():
    print('=================================================================')
    print(__intro__)
    print(__author__)
    print(__copyright__)
    print(__mod__)
    print(__disclaimer__)
    print(__cite__)
    print(__contributers__)



import os
import re
import subprocess
import numpy as np
import time as proc_time
import tensorflow.keras as keras
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Flatten
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Conv1D
from tensorflow.keras.layers import MaxPooling1D
############################################################
def KRR(Xin: np.ndarray,
        time: float,
        time_step: float,
        QDmodelIn: str,
        traj_output_file: str):
    
    QDmodelIn = re.split(r'.unf', QDmodelIn)[0]
    tm = Xin.shape[1]
    tmm = (tm-1)*time_step
    t = np.arange(0,time+tmm+ time_step, time_step)
    nsteps = len(t)
    y = Xin;
    y = np.resize(y, (y.shape[1], y.shape[0]))
    a = 1;
    #
    print('ml_dyn.KRR: Running dynamics with KRR model using MLatom in the backend ......')
    print('ml_dyn.KRR: The output of MLatom will be saved as "krr_dyn_output"')
    #
    QDmodelIn = str(os.getcwd()) + '/' + QDmodelIn + '.unf'
    ti = proc_time.time()
    for i in range(tm, len(t)):
        np.savetxt('input.dat', Xin)
        arg = ['rm', '-f', 'y_est.dat']
        subprocess.run(arg, check=True)
        args = ['mlatom', 'useMlmodel', 'MlmodelIn='+ str(QDmodelIn) + ' XfileIn=input.dat', 'YestFile=y_est.dat', 'debug']
        with open('krr_dyn_output', "w") as output:
            subprocess.run(args, check=True, stdout=output)
        y_est = np.loadtxt('y_est.dat')
        y = np.resize(y, (y.shape[0]+1, y.shape[1]))
        y[y.shape[0]-1, 0] = y_est
        Xin[0,:] = y[a:y.shape[0], 0]
        a += 1;
    arg = ['rm', '-f', 'y_est.dat']
    subprocess.run(arg, check=True)
    np.save(traj_output_file, np.c_['-1',t[:], y[:]])
    print('ml_dyn.KRR: Dynamics is saved in a file  "' + traj_output_file + '"')
    print('ml_dyn.KRR: Time taken =', proc_time.time() - ti, "sec")
    

def OSTL(Xin: np.ndarray,  
        n_states: int, 
        time: float,  
        time_step: float, 
        QDmodelIn: str,
        systemType: str,
        traj_output_file: str):
   
    QDmodelIn = str(os.getcwd()) + '/' + QDmodelIn 
    model = keras.models.load_model(QDmodelIn)
    #Show the model architecture
    model.summary()
    #
    print('ml_dyn.OSTL: Running dynamics with OSTL approach ......')
    #
    if systemType == 'FMO':
        time_range = np.arange(0,time*1000,time_step) # fs
    elif systemType == 'SB':
        time_range = np.arange(0,time,time_step) 
    time_range = np.append(time_range, time_range[time_range.shape[0]-1] + time_step, axis=None)
    nsteps = len(time_range)
    y = np.zeros((nsteps, n_states**2), dtype=float)
    y1 = np.zeros((nsteps, n_states**2), dtype=complex)
    #
    # normalizing the time feature using logistic function 
    #
    ti = proc_time.time()
    x_pred = Xin[:]
    x_pred = x_pred.reshape(1, Xin.shape[1],1) # reshape the input 
    yhat = model.predict(x_pred, verbose=0)
    a = 0; b = n_states**2;
    for i in range(0, nsteps):
        y[i,:] = yhat[0, a:b]
        a = a + n_states**2
        b = b + n_states**2
    for i in range(0, nsteps):
        # grab poplation
        a = n_states; b = n_states - 1
        c = 0; d = 0  
        for j in range(0, n_states):
            y1[i,c] = y[i,d]
            e = d + 1
            f = e + 1
            g = c + 1
            h = g + n_states - 1
            for k in range(0, b):
                y1[i,g] = y[i,e] + 1j * y[i,f]
                y1[i,h] = y[i,e] - 1j * y[i,f]
                e += 2
                f += 2
                g += 1
                h += n_states
            d += a + b
            a -= 1
            b -= 1
            c += n_states + 1
    np.save(traj_output_file, np.c_['-1',time_range, y1])
    print('ml_dyn.OSTL: Dynamics is saved in a file  "' + traj_output_file + '"')
    print('ml_dyn.OSTL: Time taken =', proc_time.time() - ti, "sec")

def AIQD(Xin: np.ndarray,  
        n_states: int, 
        time: float,  
        time_step: float, 
        numLogf: int, 
        LogCa: float,
        LogCb: float, 
        LogCc: float, 
        LogCd: float,
        QDmodelIn: str,
        systemType: str,
        traj_output_file: str):
    
    def logistic(x, c):
        
        return LogCa/(1 + LogCb * np.exp(-(x-c)/LogCd)) 
    
    QDmodelIn = str(os.getcwd()) + '/' + QDmodelIn
    model = keras.models.load_model(QDmodelIn)
    #Show the model architecture
    model.summary()
    #
    print('ml_dyn.AIQD: Running dynamics with AI-QD approach ......')
    #
    if systemType == 'FMO':
        time_range = np.arange(0,time*1000,time_step) # fs
    elif systemType == 'SB':
        time_range = np.arange(0,time,time_step) 
    time_range = np.append(time_range, time_range[time_range.shape[0]-1] + time_step, axis=None)
    if systemType == 'FMO':
        time_step = time_step/1000 # convert to ps
    t = np.arange(0,time,time_step)  
    t = np.append(t, t[t.shape[0]-1] + time_step, axis=None) # don't use time + time_step, has a bug, sometimes gives one extra value
    nsteps = len(t)
    y = np.zeros((nsteps, n_states**2), dtype=float)
    y1 = np.zeros((nsteps, n_states**2), dtype=complex)
    tt = np.zeros((nsteps, numLogf), dtype=float)
    u = 0
    for i in t:
        c = LogCc
        for j in range(0, numLogf): 
            tt[u,j]=logistic(i, c)
            c += 5.0
        u += 1
    #
    # normalizing the time feature using logistic function 
    #
    ti = proc_time.time()
    Xin_dim = Xin.shape[1]
    Xin = np.resize(Xin, (Xin.shape[0], Xin.shape[1] + numLogf))
    i = 0
    for tm in tt:
        Xin[0, Xin_dim:Xin.shape[1]] = tm[:]
        x_pred = Xin[:]
        x_pred = x_pred.reshape(1, Xin.shape[1],1) # reshape the input 
        yhat = model.predict(x_pred, verbose=0)
        y[i,:] = yhat
        i += 1
    for i in range(0, nsteps):
        # grab poplation
        a = n_states; b = n_states - 1
        c = 0; d = 0  
        for j in range(0, n_states):
            y1[i,c] = y[i,d]
            e = d + 1
            f = e + 1
            g = c + 1
            h = g + n_states - 1
            for k in range(0, b):
                y1[i,g] = y[i,e] + 1j * y[i,f]
                y1[i,h] = y[i,e] - 1j * y[i,f]
                e += 2
                f += 2
                g += 1
                h += n_states
            d += a + b
            a -= 1
            b -= 1
            c += n_states + 1
    np.save(traj_output_file, np.c_['-1',time_range, y1])
    print('ml_dyn.AIQD: Dynamics is saved in a file  "' + traj_output_file + '"')
    print('ml_dyn.AIQD: Time taken =', proc_time.time() - ti, "sec")





import os
import glob
import re
import numpy as np

def KRR(Xin: str,
        Yin: str,
        dataCol: int, 
        xlength: int, 
        dtype: str, 
        dataPath: str):
    all_files = {}
    j = 0
    print('=================================================================')
    print('prep_input.KRR: Grabbing data from "', dataPath, '" directory') 
    for files in glob.glob(dataPath+'/*'):
        file_name = os.path.basename(files)
        all_files[file_name] = np.load(files)
        j += 1
    xx = all_files[file_name]
    length = xx[:,0].shape[0]
    file_count = j
    print('prep_input.KRR: Number of trajectories =', file_count)
    m = (length - xlength) * file_count
    x = np.zeros((int(m), xlength), dtype=float)
    y = np.zeros((int(m),1), dtype=float)
    m = 0
    for files in glob.glob(dataPath+'/*'):
        file_name = os.path.basename(files)
        df = all_files[file_name]
        if dtype == 'imag':
            y1 = df[:,dataCol].imag
        else:
            y1 = df[:,dataCol].real
        for i in range(0, (length - xlength)):
            x[m,:] = y1[i:xlength+i]
            y[m,0] = y1[i+xlength]
            m += 1
        np.savetxt(Xin, x) # the input is saved as Xin
        np.savetxt(Yin, y) # the target values are saved as Yin
    print('prep_input.KRR: The input and target values are saved as (txt files), i.e.,', Xin, 'and', Yin, ',respectively')

def OSTL(Xin: str,
        Yin: str,
        systemType: str, 
        n_states: int,
        energyNorm: float,
        DeltaNorm: float,
        gammaNorm: float, 
        lambNorm: float, 
        tempNorm: float, 
        dataPath: str):
    all_files = {}
    j = 0
    print('=================================================================')
    print('prep_input.OSTL: Grabbing data from "', dataPath, '" directory')
    print('prep_input.OSTL: It is assumed that the data is in the same naming format and the same datatype as were adopted in our QD3SET-1 dataset \
, otherwise training files will not be successfully generated')
    for files in glob.glob(dataPath+'/*np[yz]'):
        file_name = os.path.basename(files)
        all_files[file_name] = np.load(files)
        j += 1
    file_count =  j
    print('prep_input.OSTL: Number of trajectories =', file_count)
    # create empty list
    gamma = np.zeros((file_count), dtype=float)
    lamb = np.zeros((file_count), dtype=float)
    temp = np.zeros((file_count), dtype=float)  # Inv-temperature (beta) in the case of SB model
    initial = np.zeros((file_count), dtype=int)
    epsilon = np.zeros((file_count), dtype=float)
    Delta = np.zeros((file_count), dtype=float)
    j = 0
    for files in glob.glob(dataPath+'/*np[yz]'):
        #
        # extract the values of gamma, lambda and temperature from the file name
        #
        file_name = os.path.basename(files)
        if systemType == 'FMO':
            x = re.split(r'_', file_name)
            y = re.split(r'-', x[1])
            initial[j] = y[1]
            y = re.split(r'-', x[2]) # extracting value of gamma
            gamma[j] = y[1] 
            y = re.split(r'-', x[3]) # extract value of lambda 
            lamb[j] = y[1]
            y = re.split(r'-', x[4])
            x = re.split(r'.npy', y[1]) # extract value of temperature
            temp[j] = x[0]
        if systemType == 'SB': 
            x = re.split(r'-', file_name)
            y = re.split(r'_', x[1])
            epsilon[j] = y[0]
            y = re.split(r'_', x[2])
            Delta[j] = y[0]
            y = re.split(r'_', x[3])
            lamb[j] = y[0]
            y = re.split(r'_', x[4])
            gamma[j] = y[0]
            y = re.split(r'.n', x[5])
            temp[j] = y[0]
        j = j + 1
#
    print('prep_input.OSTL: Normalizing gamma, lambda and temperature using the following',
            'normalizing factors in their respective order:', gammaNorm, lambNorm, tempNorm)
#
    j=0
    for i in lamb:
        lamb[j] = i/lambNorm
        j=j+1
    j=0
    for i in gamma:
        gamma[j] = i/gammaNorm
        j=j+1
    j=0
    for i in temp:
        temp[j] = i/tempNorm
        j=j+1
    j=0
    if systemType == 'SB':
        for i in epsilon:
            epsilon[j] = i/energyNorm
            j=j+1
        j=0
        for i in Delta:
            Delta[j] = i/DeltaNorm
            j=j+1

    xx = all_files[file_name]
    length = xx[:,0].shape[0]
    if systemType == 'FMO':
        x = np.zeros((file_count, 4), dtype=float)
    if systemType == 'SB':
        x = np.zeros((file_count, 5), dtype=float)
    y = np.zeros((file_count, n_states**2 * length), dtype=float)
    yy = np.zeros((1, n_states**2), dtype=complex)
    labels = np.zeros((1, int((n_states**2-n_states)/2 + n_states)), dtype = int)
    labels = []
    m = 0; 
    a = 0; b = n_states
    for i in range(0, n_states):
        for j in range(a, b):
            labels.append(j)
            m += 1
        a += n_states + 1 
        b += n_states 
    divider = n_states + 1
    m = 0
    f = 0
    for files in glob.glob(dataPath+'/*np[yz]'):
        file_name = os.path.basename(files)
        df = all_files[file_name]
        if systemType == 'FMO':
            if initial[f] == 1:   
                init_index = 0.1  # use 0.1 as a label for initial excitation on site-1
            elif initial[f] == 6:
                init_index = 0.6  # use 0.6 as a label for initial excitation on site-6
            else:
                init_index = 0.8  # use 0.8 as a lebel for initial excitation on site-8
            x[f,0] = init_index
            x[f,1] = gamma[f]
            x[f,2] = lamb[f]
            x[f,3] = temp[f]
        if systemType == 'SB':
            x[m,0] = epsilon[f]
            x[m,1] = Delta[f]
            x[m,2] = gamma[f]
            x[m,3] = lamb[f]
            x[m,4] = temp[f]
        q = 0
        for i in range(0, length):
            yy[0,:] = df[i, 1:n_states**2+1] # excluding the 1st column of time
            for p in labels:
                if p%divider == 0:
                    y[m,q] = yy[0,p].real
                    q += 1
                else:
                    y[m,q] = yy[0,p].real
                    q += 1
                    y[m,q] = yy[0,p].imag
                    q += 1
        m += 1
        f += 1
    np.save(Xin, x) # the input is saved as Xin
    np.save(Yin, y) # the target values are saved as Yin
    print('prep_input.OSTL: The input and target values are saved as', Xin + '.npy', 'and', Yin + '.npy', ', respectively.')

def AIQD(Xin: str,
        Yin: str,
        systemType: str, 
        n_states: int, 
        time: float, 
        time_step: float, 
        numLogf: int,
        LogCa: float,
        LogCb: float, 
        LogCc: float, 
        LogCd: float,
        energyNorm: float,
        DeltaNorm: float,
        gammaNorm: float, 
        lambNorm: float, 
        tempNorm:float, 
        dataPath: str):
    all_files = {}
    j = 0
    print('=================================================================')
    print('prep_input.AIQD: Grabbing data from "', dataPath, '" directory')
    print('prep_input.AIQD: It is assumed that the data is in the same naming format and the same datatype as were adopted in our QD3SET-1 dataset \
, otherwise training files will not be successfully generated')
    for files in glob.glob(dataPath+'/*np[yz]'):
        file_name = os.path.basename(files)
        all_files[file_name] = np.load(files)
        j += 1
    file_count =  j
    print('prep_input.AIQD: Number of trajectories =', file_count)
    # create empty list
    gamma = np.zeros((file_count), dtype=float)
    lamb = np.zeros((file_count), dtype=float)
    temp = np.zeros((file_count), dtype=float) # Inv-temperature (beta) in the case of SB model
    initial = np.zeros((file_count), dtype=int)
    epsilon = np.zeros((file_count), dtype=float)
    Delta = np.zeros((file_count), dtype=float)
    j = 0
    for files in glob.glob(dataPath+'/*np[yz]'):
        #
        # extract the values of gamma, lambda and temperature from the file name
        #
        file_name = os.path.basename(files)
        if systemType == 'FMO':
            x = re.split(r'_', file_name)
            y = re.split(r'-', x[1])
            initial[j] = y[1]
            y = re.split(r'-', x[2]) # extracting value of gamma
            gamma[j] = y[1] 
            y = re.split(r'-', x[3]) # extract value of lambda 
            lamb[j] = y[1]
            y = re.split(r'-', x[4])
            x = re.split(r'.npy', y[1]) # extract value of temperature
            temp[j] = x[0]
        if systemType == 'SB': 
            x = re.split(r'-', file_name)
            y = re.split(r'_', x[1])
            epsilon[j] = y[0]
            y = re.split(r'_', x[2])
            Delta[j] = y[0]
            y = re.split(r'_', x[3])
            lamb[j] = y[0]
            y = re.split(r'_', x[4])
            gamma[j] = y[0]
            y = re.split(r'.n', x[5])
            temp[j] = y[0]
        j = j + 1
#
    print('prep_input.AIQD: Normalizing gamma, lambda and temperature using the following',
            'normalizing factors in their respective order:', gammaNorm, lambNorm, tempNorm)
#
    j=0
    for i in lamb:
        lamb[j] = i/lambNorm
        j=j+1
    j=0
    for i in gamma:
        gamma[j] = i/gammaNorm
        j=j+1
    j=0
    for i in temp:
        temp[j] = i/tempNorm
        j=j+1
    if systemType == 'SB':
        j=0
        for i in epsilon:
            epsilon[j] = i/energyNorm
            j=j+1
        j=0
        for i in Delta:
            Delta[j] = i/DeltaNorm
            j=j+1

    xx = all_files[file_name]
    length = xx[:,0].shape[0]
    if systemType == 'FMO':
        time_step = time_step/1000 # convert to 5 fs
        x = np.zeros((file_count*length, 4+numLogf), dtype=float)
    if systemType == 'SB':
        x = np.zeros((file_count*length, 5+numLogf), dtype=float)
    t = np.arange(0,time, time_step)
    t = np.append(t, t[t.shape[0]-1] + time_step, axis=None)  # don't use time + time_step, has a bug, sometimes gives one extra value
    nsteps = len(t)
    y = np.zeros((nsteps, n_states**2), dtype=float)
    tt = np.zeros((nsteps, numLogf), dtype=float)
    def logistic(x, c):    
        return LogCa/(1 + LogCb * np.exp(-(x-c)/LogCd)) 
    u = 0
    for i in t:
        c = LogCc
        for j in range(0, numLogf): 
            tt[u,j]=logistic(i, c)
            c += 5.0
        u += 1
    y = np.zeros((file_count*length, n_states**2), dtype=float)
    yy = np.zeros((1, n_states**2), dtype=complex)
    labels = np.zeros((1, int((n_states**2-n_states)/2 + n_states)), dtype = int)
    labels = []
    m = 0; 
    a = 0; b = n_states
    for i in range(0, n_states):
        for j in range(a, b):
            labels.append(j)
            m += 1
        a += n_states + 1 
        b += n_states 
    divider = n_states + 1
    m = 0
    f = 0
    for files in glob.glob(dataPath+'/*np[yz]'):
        file_name = os.path.basename(files)
        df = all_files[file_name]
        for i in range(0, length):
            if systemType == 'FMO':
                if initial[f] == 1:   
                    init_index = 0.1  # use 0.1 as a label for initial excitation on site-1
                elif initial[f] == 6:
                    init_index = 0.6  # use 0.6 as a label for initial excitation on site-6
                else:
                    init_index = 0.8  # use 0.8 as a lebel for initial excitation on site-8
                x[m,0] = init_index
                x[m,1] = gamma[f]
                x[m,2] = lamb[f]
                x[m,3] = temp[f]
                x[m,4:] = tt[i,:]
            if systemType == 'SB':
                x[m,0] = epsilon[f]
                x[m,1] = Delta[f]
                x[m,2] = gamma[f]
                x[m,3] = lamb[f]
                x[m,4] = temp[f]
                x[m,5:] = tt[i,:]
            yy[0,:] = df[i, 1:n_states**2+1] # excluding the 1st column of time
            q = 0
            for p in labels:
                if p%divider == 0:
                    y[m,q] = yy[0,p].real
                    q += 1
                else:
                    y[m,q] = yy[0,p].real
                    q += 1
                    y[m,q] = yy[0,p].imag
                    q += 1
            m += 1
        f += 1
    np.save(Xin, x) # the input is saved as Xin
    np.save(Yin, y) # the target values are saved as Yin
    print('prep_input.AIQD: The input and target values are saved as', Xin + '.npy', 'and', Yin + '.npy', ', respectively.')

    
      



import os
import subprocess
import numpy as np
import mlqd.cnn as cnn
import mlqd.hyperopt_optim as optim
import time as proc_time
import tensorflow.keras as keras
import mlqd.prep_input as prep_input
############################################################
def KRR(Xin: str,
        Yin: str,
        QDmodelOut: str,
        prepInput: str, 
        dataCol: int, 
        xlength: int, 
        dtype: str, 
        dataPath: str,
        hyperParam: str,
        krrSigma: float,
        krrLamb: float):
    #
    if prepInput == 'True':
        print('train_ml.KRR: preparing training data for KRR model')
        prep_input.KRR(Xin,
                    Yin, 
                    dataCol, 
                    xlength, 
                    dtype, 
                    dataPath)
    print('=================================================================')
    print('Train_ml.KRR: Training KRR model with Gaussian kernel using MLatom in the backend ......')
    print('Train_ml.KRR: KRR model will be created and saved as "' + str(QDmodelOut) + '"')
    #
    print('Train_ml.KRR: Checking to see whether the input data files', Xin, 'and', Yin, 'exists')
    if os.path.isfile(Xin) == False:
        raise ValueError("Train_ml.KRR: file name " + str(Xin) + " does not exist")
    if os.path.isfile(Yin) == False:
        raise ValueError("Train_ml.KRR: file name ", str(Xin), " does not exist")
    print('=================================================================')
    ti = proc_time.time()
    QDmodelOut = QDmodelOut + '.unf'
    arg = ['rm', '-f', QDmodelOut + '*']
    subprocess.run(arg, check=True)
    if hyperParam == 'True':
        args = ['mlatom', 'createMLmodel', 'MLmodelOut='+ str(QDmodelOut) + ' XfileIn=' + str(Xin), 'Yfile=' + str(Yin), 
                    'kernel=Gaussian', 'sigma=opt', 'lgSigmaL=-25', 'lgSigmaH=25', 'lambda=opt', 'lgLambdaL=-30.0', 'sampling=random']
        with open('krr_train_output', "w") as output:
            subprocess.run(args, check=True, stdout=output)
    else:
        args = ['mlatom', 'createMLmodel', 'MLmodelOut='+ str(QDmodelOut) + ' XfileIn=' + str(Xin), 'Yfile=' + str(Yin), 
                    'kernel=Gaussian', 'sigma='+ str(krrSigma), 'lambda=' + str(krrLamb)]
        with open('krr_train_output', "w") as output:
            subprocess.run(args, check=True, stdout=output)
    print('Train_ml.KRR: The output of MLatom can be found as "krr_train_output" file',
    '(please check to ensure that MLatom execution was successful)') 
    print('Train_ml.KRR: Time taken =', proc_time.time() - ti, "sec")

def OSTL(Xin: str,
        Yin: str, 
        systemType: str,
        n_states: int,
        energyNorm: float,
        DeltaNorm: float,
        gammaNorm: float,
        lambNorm: float, 
        tempNorm: float,
        dataPath: str,  # optional in prepInput is False
        QDmodelOut: str,
        hyperParam: str,
        OptEpochs: int,
        TrEpochs: int,
        max_evals: int, 
        patience: int,
        prepInput: str):

    if prepInput == 'True':
        print('train_ml.OSTL: preparing training data for OSTL model')
        prep_input.OSTL(Xin,
                        Yin,
                        systemType, 
                        n_states,
                        energyNorm,
                        DeltaNorm,
                        gammaNorm, 
                        lambNorm, 
                        tempNorm, 
                        dataPath)

    if hyperParam == 'True':
        print('=================================================================')
        print('Train_ml.AIQD: Going for hyperopt optimization of CNN')
        print('=================================================================')
        t1 = proc_time.time()
        optim.optimize(Xin, Yin, OptEpochs, max_evals)
        print('Train_ml.OSTL: Time taken for optimization =', proc_time.time() - t1, "sec")
        print('Train_ml.OSTL: Training CNN model with the optimized hyper_parameters')
        print('=================================================================')
        t2 = proc_time.time()
        cnn.CNN_optim(Xin, Yin, QDmodelOut, TrEpochs, patience)
        print('Train_ml.OSTL: Time taken for training =', proc_time.time() - t2, "sec")
        print('Train_ml.OSTL: Total Time (optimization + training) =', proc_time.time() - t1 , "sec")
    else:
        optim_param_file = "best_param.pkl"
        print('=================================================================')
        print('Train.ml_OSTL: Looking for',optim_param_file)
        print('=================================================================')
        t1 = proc_time.time()
        if os.path.isfile(optim_param_file):
            print('Train.ml_OSTL: loading hyperparameters from', optim_param_file)
            cnn.CNN_optim(Xin, Yin, QDmodelOut, TrEpochs, patience)
            print('Train_ml.OSTL: Time taken for training =', proc_time.time() - t1 , "sec")
        else:
            print('=================================================================')
            print('Train.ml_OSTL: '+ str(optim_param_file) +  ' not found, thus training CNN model with the default structure')
            print('=================================================================')
            cnn.OSTL_default(Xin, Yin, QDmodelOut, TrEpochs, patience)
            print('Train_ml.OSTL: Time taken for training =', proc_time.time() - t1 , "sec")

def AIQD(Xin: str,
        Yin: str,
        systemType: str, 
        n_states: int, 
        time: float, 
        time_step: float, 
        numLogf: int,
        LogCa: float,
        LogCb: float, 
        LogCc: float, 
        LogCd: float,
        energyNorm: float,
        DeltaNorm: float,
        gammaNorm: float, 
        lambNorm: float, 
        tempNorm: float,
        QDmodelOut: str,
        hyperParam: str,
        OptEpochs: int,
        TrEpochs: int, 
        max_evals: int, 
        patience: int,
        prepInput: str,
        dataPath: str):

    if prepInput == 'True':
        print('train_ml.AIQD: preparing training data for AIQD model')
        prep_input.AIQD(
                        Xin,
                        Yin,
                        systemType, 
                        n_states,
                        time,
                        time_step, 
                        numLogf,
                        LogCa,
                        LogCb, 
                        LogCc, 
                        LogCd,
                        energyNorm,
                        DeltaNorm,
                        gammaNorm, 
                        lambNorm, 
                        tempNorm, 
                        dataPath
                        )

    if hyperParam == 'True':
        print('=================================================================')
        print('Train_ml.AIQD: Going for hyperopt optimization of CNN')
        print('=================================================================')
        t1 = proc_time.time()
        optim.optimize(Xin, Yin, OptEpochs, max_evals)
        print('Train_ml.OSTL: Time taken for optimization =', proc_time.time() - t1, "sec")
        print('Train_ml.AIQD: Training CNN model with the optimized hyper_parameters')
        print('=================================================================')
        t2 = proc_time.time()
        cnn.CNN_optim(Xin, Yin, QDmodelOut, TrEpochs, patience)
        print('Train_ml.AIQD: Time taken for training =', proc_time.time() - t2, "sec")
        print('Train_ml.AIQD: Total Time (optimization + training) =', proc_time.time() - t1 , "sec")

    else:
        optim_param_file = "best_param.pkl"
        print('=================================================================')
        print('Train.ml_AIQD: Looking for',optim_param_file)
        print('=================================================================')
        t1 = proc_time.time()
        if os.path.isfile(optim_param_file):
            print('Train.ml_AIQD: loading hyperparameters from', optim_param_file)
            cnn.CNN_optim(Xin, Yin, QDmodelOut, TrEpochs, patience)
            print('Train_ml.AIQD: Time taken for training =', proc_time.time() - t1, "sec")
        else:
            print('=================================================================')
            print('Train.ml_OSTL: '+ str(optim_param_file) +  ' not found, thus training CNN model with the default structure')
            print('=================================================================')
            cnn.AIQD_default(Xin, Yin, QDmodelOut, TrEpochs,  patience)
            print('Train_ml.AIQD: Time taken for training =', proc_time.time() - t1, "sec")



import os
import numpy as np
from sklearn.model_selection import train_test_split; 
def data(Xin:str, Yin: str):

    Xin = Xin + '.npy'
    Yin = Yin + '.npy'
    print('Data: Checking to see whether the input data files', Xin, 'and', Yin, 'exist')
    if os.path.isfile(Xin) == False:
        raise ValueError("Data: file name " + str(Xin) + " does not exist")
    if os.path.isfile(Yin) == False:
        raise ValueError("Data: file name " + str(Yin) + " does not exist")
    
    print('Data: Loading data files', Xin, Yin)
    pwd = os.getcwd()
    X = str(pwd) + '/' + str(Xin)
    Y = str(pwd) + '/' + str(Yin)
    x = np.load(X)
    y = np.load(Y)
    print('Data: splitting data into 80/20 % ratio (training/validation set)')
    x_train, x_val, y_train, y_val = train_test_split(x, y, test_size=0.2, random_state=7)
    x_train = x_train.reshape(x_train.shape[0], x_train.shape[1],1)
    x_val  = x_val.reshape(x_val.shape[0], x_val.shape[1],1)
    kernel_choice = []
    for i in range(2, x_train.shape[1], 1):
        kernel_choice.append(i)
    return x_train, y_train, x_val, y_val, kernel_choice




import os
import re
import numpy as np
import matplotlib.pyplot as plt

def plot(n_states: int, 
	    QDmodelType: int,
        QDtrajOut: str, 
        refTraj: str, 
        dataCol: int, 
        dtype: str,
        pltNstates: int,
        xlim: int):
    QDtrajOut = re.split(r'.npy', QDtrajOut)[0]
    X = str(os.getcwd()) + "/" + str(QDtrajOut) + ".npy" 
    pred = np.load(X)
    ref = np.load(refTraj)
    if QDmodelType == 'KRR':
        if dtype == 'real':
            plt.plot(np.real(pred[:,0]), np.real(pred[:,dataCol]), lw=2.0, label=str(dataCol))
            plt.plot(np.real(ref[:,0]), np.real(ref[:,dataCol]), lw=1.0, ls='--', marker='o', markevery=20, label='_nolegend_')
            plt.legend(bbox_to_anchor=(0.5, 1.1), loc="upper center", ncol=n_states, frameon=False)
        elif dtype == 'imag':
            plt.plot(np.real(pred[:,0]), np.imag(pred[:,dataCol]), lw=2.0, label=str(dataCol))
            plt.plot(np.real(ref[:,0]), np.imag(ref[:,dataCol]), lw=1.0, ls='--', marker='o', label='_nolegend_')
            plt.legend(bbox_to_anchor=(0.5, 1.1), loc="upper center", ncol=n_states, frameon=False)
        plt.xlabel("time")
        if xlim is not None:
            plt.xlim(0, xlim)
        plt.savefig('krr_plot.png', bbox_inches='tight')
        print('Dynamics is plotted and saved as "krr_plot.png". Dash Line with markers is the reference trajectory while the solid line is the predicted dynamics')
        plt.close()
    else:
        a=1
        for i in range(1, pltNstates**2 + n_states, n_states+1):
            idx = str(a) + str(a) 
            plt.plot(np.real(pred[:,0]), np.real(pred[:,i]), lw=2.0, label=idx)
            plt.plot(np.real(ref[:,0]), np.real(ref[:,i]), lw=1.0, ls='--', marker='o',markevery=20, label='_nolegend_')
            plt.legend(bbox_to_anchor=(0.5, 1.1), loc="upper center", ncol=n_states, frameon=False)
            a += 1
        plt.xlabel("time")
        if xlim is not None:
            plt.xlim(0, xlim)
        plt.savefig('population.png', bbox_inches='tight')
        print('The diagonal part (population) is plotted and saved as "population.png". Dash Line with markers is the reference trajectory while the solid line is the predicted dynamics')
        plt.close()
        a = 2; b = n_states
        c = 1;  
        for j in range(0, pltNstates):
            d = c + 1 
            for i in range(a, b + 1): 
                idx = str(c) + str(d)
                plt.plot(np.real(pred[:,0]), np.real(pred[:,i]), lw=2.0, label=idx)
                plt.plot(np.real(ref[:,0]), np.real(ref[:,i]), lw=1.0, ls='--', marker='o', markevery=20, label='_nolegend_')
                plt.legend(bbox_to_anchor=(0.5, 1.1), loc="upper center", ncol=n_states, frameon=False)
                d += 1
            a += n_states + 1
            b += n_states
            c += 1
        plt.xlabel("time")
        if xlim is not None:
            plt.xlim(0, xlim)
        plt.savefig('real_part_coherence.png', bbox_inches='tight')
        print('The real part of the offdiagonal elements is plotted and saved as "real_part_coherence.png". Dash Line with markers is the reference trajectory while the solid line is the predicted dynamics ')
        plt.close()
        a = 2; b = n_states
        c = 1;  
        for j in range(0, pltNstates):
            d = c + 1 
            for i in range(a, b + 1): 
                idx = str(c) + str(d)
                plt.plot(np.real(pred[:,0]), np.imag(pred[:,i]), lw=2.0, label=idx)
                plt.plot(np.real(ref[:,0]), np.imag(ref[:,i]), lw=1.0, ls='--', marker='o',markevery=20, label='_nolegend_')
                plt.legend(bbox_to_anchor=(0.5, 1.1), loc="upper center", ncol=n_states, frameon=False)
                d += 1
            a += n_states + 1
            b += n_states
            c += 1
        plt.xlabel("time")
        if xlim is not None:
            plt.xlim(0, xlim)
        plt.savefig('imag_part_coherence.png', bbox_inches='tight')
        print('The imaginary part of the offdiagonal elements is plotted and saved as "imag_part_coherence.png". Dash Line with markers is the reference trajectory while the solid line is the predicted dynamics ')
        plt.close()




__initState__='initState (int): Initial state with Initial Excitation. It is only required when propagating dynamics with OSTL or AIQD method for FMO complex. Default value is 1'
__n_states__='n_states (int): The number of states or sites. Default values are 2 for spin-boson model and 7 for FMO complex'
__QDmodel__='QDmodel (str): You can pass "createQDmodel" to train a QD model or "useQDmodel" to propagate dynamics with the already trained QD model. Default option is "useQDmodel"'
__systemType__='systemType (str): System type, you can pass "SB" for spin-boson model and "FMO" for FMO complex'
__QDmodelOut__='QDmodelOut (str): MLQD will save the trained QD model with this name. If you don not pass it, MLQD will pick a random name'
__QDmodelIn__='QDmodelIn (str): Pass the name of the trained QD model and MLQD will use it to predict dynamics'
__QDtrajOut__='QDtrajOut (str): MLQD will save the predicted trajectory with this name. If you do not pass a name, MLQD will pick a random name'
__QDmodelType__='QDmodelType (str): Type of QD model. You can pass "KRR" for kernel ridge regression method, "OSTL" for OSTL approach and "AIQD" for AI-QD approach. The default option is OSTL'
__prepInput__='prepInput (str): Choose wether you wanna prepare the input X and Y files from the data. You can pass "True" or "False". Default option is "False". For OSTL and AIQD, your data files should be in the same naming and file format as in out QD3SET-1 dataset' 
__time__='time: (float): Propagation time in picoseconds (ps) for FMO complex and in atomic units (a.u.) for spin-boson model'
__time_step__='time_step (float): Time-step for time-propagation. Default values are 0.05 a.u. (for spin-boson model) and 5 fs for FMO complex'
__hyperParam__='hyperParam (str): Default option is "False". You can pass "True" (to optimize the hyperparameters) or "False" to not optimize the hyperparameters and run with the default structure'
__patience__='patience (int): Patience for early stopping in CNN training'
__OptEpochs__='OptEpochs (int): The number of epochs for optimzation. Default value is 100'
__TrEpochs__='TrEpochs (int): The number of epochs for training. Default value is 100'
__max_evals='max_evals (int): The number of maximum evaluations in hyperopt optimization. Default value is 100'
__XfileIn__='XfileIn (str): The prepared X file will be saved at the provided file name, if not provided, MLQD will use the default name. In the case of KRR-based prediction of dynamics, this option provides the input short-time dynamics'  
__YfileIn__='YfileIn (str): The prepared Y file will be saved at the provided file name. If not provided, MLQD will use the dafault name' 
__datapath__='dataPath (str): MLQD will access data with this path and prepare the X and Y files'
__numLogf__='numLogf (int): The number of Logistic functions for the normalization of time dimension. Default value is 1.0'    
__LogCa__='LogCa (float): Coefficient "a" in the logistic function, default value is 1.0'
__LogCb__='LogCb (float): Coefficient "b" in the logistic function, default value is 15.0'
__LogCc__='LogCc (float): Coefficient "c" in the logistic function, default value is -1.0'
__LogCd__='LogCd (float): Coefficient "d" in the logistic function, default values is 1.0'
__energyDiff__='energyDiff (float): Energy difference between the two states in the case of spin-boson model. Default value is 1.0'
__Delta__='Delta (float): The tunneling matrix element in the case of spin-boson model. Default value is 1.0' 
__gamma__='gamma (int): Characteristic frequency. Default values are 10 (spin-boson model) and 500 (FMO complex)'
__lamb__='lamb (int): System-bath coupling strength or the reorganization energy. Default values are 1 (spin-boson model) and 520 (FMO complex)'
__temp__='temp (int): Temperature or inverse temperature. Default values are 1 (spin-boson model) and 510 (FMO complex)' 
__energyNorm__='energyNorm (float): Normalizer for the energy difference between the states in the case of spin-boson model. Default value is 1.0'
__DeltaNorm__='DeltaNorm (float): Normalizer for the tunneling matrix element in the case of spin-boson model. Default value is 1.0'
__gammaNorm__='gammaNorm (float): Normalizer for Characteristic frequency. Default values are 500 (FMO complex) and 10 (spin-boson model)' 
__lambNorm__='lambNorm (float): Normalizer for System-bath coupling strength. Default values are 520 (FMO complex) and 1 (SB model)' 
__tempNorm__='tempNorm (float): Normalizer for temperature or inverse temperature. Default values are 510 (FMO complex) and 1 (spin-boson model)'
__xlim__='xlim (float): The xaxis limit while plotting' 
__pltNstates__='pltNstates (int): The number of states to be plotted. Default option is to plot all states'
__refTraj__='refTraj (str): MLQD will plot the predicted dynamics against this reference trajectory. If not provided, MLQD will ignore plotting'
_dataCol__='dataCol (int): The column number (counting starts from 0); As KRR is a single output model, and it is possible that you have many columns in your data files, using dataCol option, MLQD is able to grab only the mentioned column and train a KRR model on it. For plotting, the same option is used to grab the concerend column from the reference trajectory'
__dtype__='dtype (str): specify the data type. You can pass "real" or "imag". Default is "real". In KRR training, MLQD will extract the real or imaginary (imag) part of the column passed with "dataCol". For plotting, the same option is used the grab the concerned data type from the column of the reference trajectory'
__xlength__='xlength (int): Length of the input short-time trajectory. It is the number of time steps in the data you pass with the "dataCol". Default value is 81'
__krrSigma__='krrSigma (float): If we pass "False" to hyperParam, then we need to provide a value for the hyperparameter Sigma in Gaussian kernel. Otherwise the model will run with the default value 4.0'
__krrLamb__='krrLamb (float): If we pass "False" to hyperParam, then we need to provide a value for the hyperparameter Lambda in KRR. Otherwise the model will run with the default value 0.00000001'
def help():
    print('\n=================================================================')
    print('                            Manual                               ')
    print('=================================================================')
    i=0
    i+=1; print(str(i) + ']===>',__QDmodel__)
    i+=1; print(str(i) + ']===>',__systemType__)
    i+=1; print(str(i) + ']===>',__QDmodelOut__)
    i+=1; print(str(i) + ']===>',__QDmodelIn__)
    i+=1; print(str(i) + ']===>',__QDtrajOut__)
    i+=1; print(str(i) + ']===>',__QDmodelType__)
    i+=1; print(str(i) + ']===>',__prepInput__)
    i+=1; print(str(i) + ']===>',__time__)
    i+=1; print(str(i) + ']===>',__time_step__)
    i+=1; print(str(i) + ']===>', __hyperParam__)
    i+=1; print(str(i) + ']===>', __XfileIn__)
    i+=1; print(str(i) + ']===>', __YfileIn__)
    i+=1; print(str(i) + ']===>', __datapath__)
    print('\n------------------------------------------------------------------')
    print('                Specific to AIQD and OSTL approach                ')
    print('------------------------------------------------------------------')
    i+=1; print(str(i) + ']===>', __initState__)
    i+=1; print(str(i) + ']===>', __n_states__)
    i+=1; print(str(i) + ']===>', __patience__)
    i+=1; print(str(i) + ']===>', __OptEpochs__)
    i+=1; print(str(i) + ']===>', __TrEpochs__)
    i+=1; print(str(i) + ']===>', __max_evals)
    i+=1; print(str(i) + ']===>', __energyDiff__)
    i+=1; print(str(i) + ']===>', __Delta__)
    i+=1; print(str(i) + ']===>', __gamma__)
    i+=1; print(str(i) + ']===>', __lamb__) 
    i+=1; print(str(i) + ']===>', __temp__) 
    i+=1; print(str(i) + ']===>',__energyNorm__)
    i+=1; print(str(i) + ']===>',__DeltaNorm__)
    i+=1; print(str(i) + ']===>', __gammaNorm__)
    i+=1; print(str(i) + ']===>', __lambNorm__)
    i+=1; print(str(i) + ']===>', __tempNorm__)
    print('\n------------------------------------------------------------------')
    print('                     Specific to AIQD approach                    ')
    print('------------------------------------------------------------------')
    i+=1; print(str(i) + ']===>', __numLogf__)
    i+=1; print(str(i) + ']===>', __LogCa__)
    i+=1; print(str(i) + ']===>', __LogCb__)
    i+=1; print(str(i) + ']===>', __LogCc__)
    i+=1; print(str(i) + ']===>', __LogCd__)
    print('\n------------------------------------------------------------------')
    print('                    Specific to KRR approach        ')
    print('------------------------------------------------------------------')
    i+=1; print(str(i) + ']===>',__xlength__)
    i+=1; print(str(i) + ']===>',__krrSigma__)
    i+=1; print(str(i) + ']===>',__krrLamb__)
    print('\n------------------------------------------------------------------')
    print('                 Specific to KRR approach and Potting  ')
    print('------------------------------------------------------------------')
    i+=1; print(str(i) + ']===>', _dataCol__)
    i+=1; print(str(i) + ']===>', __dtype__)
    print('\n------------------------------------------------------------------')
    print('                        Specific to Plotting         ')
    print('------------------------------------------------------------------')
    i+=1; print(str(i) + ']===>', __xlim__) 
    i+=1; print(str(i) + ']===>', __pltNstates__)
    i+=1; print(str(i) + ']===>', __refTraj__)
    print('=================================================================')

